import json
from datetime import datetime
from typing import Any, Optional
from uuid import UUID, uuid4

from pydantic import Field, field_validator, model_validator

from whiskerrag_types.model.timeStampedModel import TimeStampedModel


class Tenant(TimeStampedModel):
    tenant_id: str = Field(
        default_factory=lambda: str(uuid4()), description="tenant id"
    )
    tenant_name: str = Field("", description="tenant name")
    email: str = Field(..., description="email")
    secret_key: str = Field("", description="secret_key")
    is_active: bool = Field(True, description="is active")
    metadata: Optional[dict] = Field(
        None, description="Metadata for the tenant", alias="metadata"
    )

    def update(self, **kwargs: Any) -> "Tenant":
        for key, value in kwargs.items():
            setattr(self, key, value)
        self.updated_at = datetime.now()
        return self

    @field_validator("is_active", mode="before")
    @classmethod
    def convert_tinyint_to_bool(cls, v: Any) -> bool:
        return bool(v)

    @model_validator(mode="before")
    def pre_process_data(cls, data: dict) -> dict:
        for field, value in data.items():
            if isinstance(value, UUID):
                data[field] = str(value)
            if field == "metadata" and isinstance(value, str):
                data[field] = json.loads(value)
        return data



from abc import ABC, abstractmethod
from typing import Generic, List, TypeVar

from whiskerrag_types.model.knowledge import Knowledge
from whiskerrag_types.model.multi_modal import Image, Text

T = TypeVar("T", Image, Text)


class BaseLoader(ABC, Generic[T]):
    def __init__(self, knowledge: Knowledge):
        self.knowledge = knowledge

    @abstractmethod
    async def load(self) -> List[T]:
        pass



import re
from typing import List, Literal, Optional, Union

from pydantic import BaseModel, Field, field_validator, model_validator


class BaseSplitConfig(BaseModel):
    """Base split configuration class"""

    chunk_size: int = Field(default=1500, ge=1, description="chunk max size")
    chunk_overlap: int = Field(
        default=150,
        ge=0,
        description="chunk overlap size, must be less than chunk_size",
    )

    @model_validator(mode="after")
    def validate_config(self) -> "BaseSplitConfig":
        if self.chunk_overlap >= self.chunk_size:
            raise ValueError("chunk_overlap must be less than chunk_size")
        return self


class BaseCharSplitConfig(BaseSplitConfig):
    """Base char split configuration class"""

    separators: Optional[List[str]] = Field(
        default=None, description="separator list, if None, use default separators"
    )
    split_regex: Optional[str] = Field(
        default=None, description="split_regex,if set, use it instead of separators"
    )

    @field_validator("split_regex")
    @classmethod
    def validate_regex(cls, v: Optional[str]) -> Optional[str]:
        if v is not None:
            try:
                re.compile(v)
            except re.error as e:
                raise ValueError(f"Invalid regular expression: {str(e)}")
        return v


class MarkdownSplitConfig(BaseSplitConfig):
    type: Literal["markdown"] = "markdown"
    separators: List[str] = Field(
        ...,
        description="""List of separators to split the text. If None, uses default separators""",
    )
    is_separator_regex: bool = Field(
        ...,
        description="""If true, the separators should be in regular expression format.""",
    )
    keep_separator: Optional[Union[bool, Literal["start", "end"]]] = Field(
        default=False,
        description="""Whether to keep the separator and where to place it in each corresponding chunk (True='start')""",
    )
    extract_header_first: Optional[bool] = Field(
        default=False,
        description="""If true, will split markdown by header first""",
    )
    # TODO :extract images,table,code


class PDFSplitConfig(BaseSplitConfig):
    """PDF document split configuration"""

    type: Literal["pdf"] = "pdf"
    extract_images: bool = Field(default=False, description="Whether to extract images")
    table_extract_mode: str = Field(
        default="text", description="Table extraction mode: 'text' or 'structure'"
    )


class TextSplitConfig(BaseCharSplitConfig):
    """Plain text split configuration"""

    type: Literal["text"] = "text"
    separators: List[str] = Field(
        ...,
        description="""List of separators to split the text. If None, uses default separators""",
    )
    is_separator_regex: bool = Field(
        ...,
        description="""If true, the separators should be in regular expression format.""",
    )
    keep_separator: Optional[Union[bool, Literal["start", "end"]]] = Field(
        default=False,
        description="""Whether to keep the separator and where to place it in each corresponding chunk (True='start')""",
    )


class JSONSplitConfig(BaseModel):
    """
    JSON document split configuration
    @link {https://python.langchain.com/api_reference/text_splitters/json/langchain_text_splitters.json.RecursiveJsonSplitter.html}
    """

    type: Literal["json"] = "json"
    max_chunk_size: int = Field(
        default=2000,
        description=""" The maximum size for each chunk. Defaults to 2000 """,
    )
    min_chunk_size: Optional[int] = Field(
        default=200,
        description="""The minimum size for a chunk. If None,
                defaults to the maximum chunk size minus 200, with a lower bound of 50.""",
    )


class YuqueSplitConfig(BaseSplitConfig):
    type: Literal["yuque"] = "yuque"
    separators: List[str] = Field(
        ...,
        description="""List of separators to split the text. If None, uses default separators""",
    )
    is_separator_regex: bool = Field(
        ...,
        description="""If true, the separators should be in regular expression format.""",
    )


class GeaGraphSplitConfig(BaseModel):
    """
    JSON document split configuration
    @link {https://python.langchain.com/api_reference/text_splitters/json/langchain_text_splitters.json.RecursiveJsonSplitter.html}
    """

    type: Literal["geagraph"] = "geagraph"
    kisId: int = Field(
        ...,
        description=""" The Kis platform business id  """,
    )


class ImageSplitConfig(BaseModel):
    type: Literal["image"] = "image"



from typing import List

from whiskerrag_types.interface import BaseDecomposer
from whiskerrag_types.model.knowledge import (
    GithubRepoSourceConfig,
    Knowledge,
    KnowledgeSourceEnum,
    KnowledgeTypeEnum,
    YuqueSourceConfig,
)
from whiskerrag_utils.helper.yuque import ExtendedYuqueLoader
from whiskerrag_utils.loader.github.repo_loader import (
    GitFileElementType,
    GithubRepoLoader,
)
from whiskerrag_utils.registry import RegisterTypeEnum, register


@register(RegisterTypeEnum.DECOMPOSER, KnowledgeTypeEnum.FOLDER)
class FolderDecomposer(BaseDecomposer):
    async def decompose(self) -> List[Knowledge]:
        results = []
        if self.knowledge.source_type == KnowledgeSourceEnum.GITHUB_REPO:
            results = await self.get_knowledge_list_from_github_repo(self.knowledge)
        if self.knowledge.source_type == KnowledgeSourceEnum.YUQUE:
            results = await self.get_knowledge_list_from_yuque(self.knowledge)
        return results

    async def get_knowledge_list_from_github_repo(
        self,
        knowledge: Knowledge,
    ) -> List[Knowledge]:
        if isinstance(knowledge.source_config, GithubRepoSourceConfig):
            repo_name = knowledge.source_config.repo_name
        else:
            raise TypeError(
                "source_config must be of type GithubRepoSourceConfig to access repo_name"
            )
        auth_info = knowledge.source_config.auth_info
        branch_name = knowledge.source_config.branch
        github_loader = GithubRepoLoader(repo_name, branch_name, auth_info)
        file_list: List[GitFileElementType] = github_loader.get_file_list()
        github_repo_list: List[Knowledge] = []
        for file in file_list:
            if not file.path.endswith((".md", ".mdx")):
                continue
            else:
                child_knowledge = Knowledge(
                    **knowledge.model_dump(
                        exclude={
                            "source_type",
                            "knowledge_type",
                            "knowledge_name",
                            "source_config",
                            "tenant_id",
                            "file_size",
                            "file_sha",
                            "metadata",
                            "parent_id",
                            "enabled",
                        }
                    ),
                    source_type=KnowledgeSourceEnum.GITHUB_FILE,
                    knowledge_type=KnowledgeTypeEnum.MARKDOWN,
                    knowledge_name=f"{file.repo_name}/{file.path}",
                    source_config={
                        **knowledge.source_config.model_dump(),
                        "path": file.path,
                    },
                    tenant_id=knowledge.tenant_id,
                    file_size=file.size,
                    file_sha=file.sha,
                    metadata={},
                    parent_id=knowledge.knowledge_id,
                    enabled=True,
                )
                github_repo_list.append(child_knowledge)
        return github_repo_list

    async def get_knowledge_list_from_yuque(
        self,
        knowledge: Knowledge,
    ) -> List[Knowledge]:
        source_config = knowledge.source_config
        if not isinstance(source_config, YuqueSourceConfig):
            raise TypeError("source_config must be of type YuqueSourceConfig")
        knowledge_list: List[Knowledge] = []
        document_id = source_config.document_id
        group_login = source_config.group_login
        book_slug = source_config.book_slug
        loader = ExtendedYuqueLoader(
            access_token=source_config.auth_info,
            api_url=source_config.api_url,
        )
        try:
            # Case 1: If document_id is provided, create single knowledge
            if document_id:
                if not book_slug:
                    raise ValueError("book_id is required when document_id is provided")
                doc_detail = loader.get_doc_detail(group_login, book_slug, document_id)
                knowledge_list.append(
                    Knowledge(
                        space_id=knowledge.space_id,
                        source_type=KnowledgeSourceEnum.YUQUE,
                        knowledge_type=KnowledgeTypeEnum.YUQUEDOC,
                        knowledge_name=doc_detail["title"],
                        split_config=knowledge.split_config.model_dump(),
                        source_config=source_config,
                        tenant_id=knowledge.tenant_id,
                        file_size=doc_detail.get(
                            "word_count"
                        ),  # You might want to calculate actual size
                        file_sha="",  # You might want to calculate actual sha
                        metadata={
                            "document_id": doc_detail.get("id"),
                            "format": doc_detail.get("format"),
                            "word_count": doc_detail.get("word_count"),
                        },
                        parent_id=knowledge.knowledge_id,
                        enabled=True,
                    )
                )

            # Case 2: If only book_id is provided, create knowledge for each document in the book
            elif book_slug:
                book_toc = loader.get_book_documents_by_path(group_login, book_slug)
                for doc_detail in book_toc:
                    doc_id = doc_detail.get("slug")
                    if doc_id is None:
                        continue
                    knowledge_list.append(
                        Knowledge(
                            space_id=knowledge.space_id,
                            source_type=KnowledgeSourceEnum.YUQUE,
                            knowledge_type=KnowledgeTypeEnum.YUQUEDOC,
                            knowledge_name=doc_detail.get("title", doc_id),
                            source_config=YuqueSourceConfig(
                                api_url=source_config.api_url,
                                group_login=group_login,
                                book_slug=book_slug,
                                document_id=doc_id,
                                auth_info=source_config.auth_info,
                            ),
                            split_config=knowledge.split_config.model_dump(),
                            tenant_id=knowledge.tenant_id,
                            file_size=0,
                            file_sha="",
                            metadata={"document_id": doc_id},
                            parent_id=knowledge.knowledge_id,
                            enabled=True,
                        )
                    )

            # Case 3: If only group_id is provided, get all books and their documents
            else:
                books = loader.get_books(user_id=loader.get_user_id())
                for book in books:
                    book_slug = book.get("slug")
                    if book_slug is None:
                        raise Exception(
                            f"can not get book slug   from knowledge:{knowledge.source_config}"
                        )
                    book_toc = loader.get_book_documents_by_path(group_login, book_slug)
                for doc_detail in book_toc:
                    doc_id = doc_detail.get("slug")
                    if doc_id is None:
                        continue
                    knowledge_list.append(
                        Knowledge(
                            space_id=knowledge.space_id,
                            source_type=KnowledgeSourceEnum.YUQUE,
                            knowledge_type=KnowledgeTypeEnum.YUQUEDOC,
                            knowledge_name=doc_detail.get("title", doc_id),
                            source_config=YuqueSourceConfig(
                                api_url=source_config.api_url,
                                group_login=group_login,
                                book_slug=book_slug,
                                document_id=doc_id,
                                auth_info=source_config.auth_info,
                            ),
                            split_config=knowledge.split_config.model_dump(),
                            tenant_id=knowledge.tenant_id,
                            file_size=0,
                            file_sha="",
                            metadata={"document_id": doc_id},
                            parent_id=knowledge.knowledge_id,
                            enabled=True,
                        )
                    )

            return knowledge_list

        except Exception as e:
            raise Exception(f"Failed to get knowledge list from Yuque: {e}")



import os
from typing import List, Optional

from langchain_openai import OpenAIEmbeddings

from whiskerrag_types.interface.embed_interface import BaseEmbedding, Image
from whiskerrag_types.model.knowledge import EmbeddingModelEnum
from whiskerrag_utils import RegisterTypeEnum, register


@register(RegisterTypeEnum.EMBEDDING, EmbeddingModelEnum.OPENAI)
class OpenAIEmbedding(BaseEmbedding):
    @classmethod
    async def health_check(cls) -> bool:
        try:
            if not os.getenv("OPENAI_API_KEY"):
                raise EnvironmentError(
                    "OPENAI_API_KEY is not set in the environment variables"
                )
            return True
        except Exception as e:
            print(f"OpenAIEmbedding health check failed: {e}")
            return False

    async def embed_text(self, text: str, timeout: Optional[int]) -> List[float]:
        embedding_client = OpenAIEmbeddings(timeout=timeout or 15)
        embedding = embedding_client.embed_query(text)
        return embedding

    async def embed_image(self, image: Image, timeout: Optional[int]) -> List[float]:
        raise NotImplementedError("OpenAI does not support image embedding")
        return []



from abc import ABC, abstractmethod
from typing import List

from whiskerrag_types.model.knowledge import Knowledge


class BaseDecomposer(ABC):
    def __init__(self, knowledge: Knowledge):
        self.knowledge = knowledge

    @abstractmethod
    async def decompose(self) -> List[Knowledge]:
        pass



from typing import Dict, Iterator, List, Union

from langchain_community.document_loaders.yuque import YuqueLoader
from langchain_core.documents import Document


class ExtendedYuqueLoader(YuqueLoader):
    """Extended Yuque loader with additional API support."""

    def __init__(
        self,
        access_token: str,
        api_url: str = "https://www.yuque.com",
    ):
        super().__init__(access_token, api_url)

    @property
    def headers(self) -> Dict[str, str]:
        return {
            "Content-Type": "application/json",
            "X-Auth-Token": self.access_token,
        }

    def get_book_toc(self, group_login: str, book_slug: str) -> Dict:
        """Get table of contents for a book."""
        url = f"{self.api_url}/api/v2/repos/{group_login}/{book_slug}/toc"
        response = self.http_get(url=url)
        return response["data"]  # type: ignore

    def get_book_detail(self, group_login: str, book_slug: str) -> Dict:
        """Get detailed information about a book."""
        url = f"{self.api_url}/api/v2/repos/{group_login}/{book_slug}"
        response = self.http_get(url=url)
        return response["data"]  # type: ignore

    def get_doc_detail(
        self, group_login: str, book_slug: str, doc_id: Union[str, int]
    ) -> Dict:
        """Get detailed information about a document."""
        url = f"{self.api_url}/api/v2/repos/{group_login}/{book_slug}/docs/{doc_id}"
        response = self.http_get(url=url)
        return response["data"]  # type: ignore

    def parse_document(self, document: Dict) -> Document:
        """Parse document into Document format."""
        content = self.parse_document_body(document.get("body", ""))
        metadata = {
            "id": document.get("id"),
            "title": document.get("title"),
            "description": document.get("description"),
            "created_at": document.get("created_at"),
            "updated_at": document.get("updated_at"),
            "published_at": document.get("published_at"),
            "word_count": document.get("word_count"),
            "book_id": document.get("book_id"),
            "user_id": document.get("user_id"),
        }
        return Document(page_content=content, metadata=metadata)

    def get_book_documents_by_path(
        self, group_login: str, book_slug: str
    ) -> List[dict]:
        """get all documents from a specific book."""
        # Get book TOC
        toc = self.get_book_toc(group_login, book_slug)

        # Filter only DOC type entries
        doc_entries = [entry for entry in toc if entry["type"] == "DOC"]
        return doc_entries

    def load_book_by_path(self, group_login: str, book_slug: str) -> Iterator[Document]:
        """Load all documents from a specific book."""
        doc_entries = self.get_book_documents_by_path(group_login, book_slug)
        for entry in doc_entries:
            doc_id = entry["doc_id"]
            doc_detail = self.get_doc_detail(group_login, book_slug, doc_id)
            yield self.parse_document(doc_detail)

    def load_document_by_path(
        self, group_login: str, book_slug: str, document_id: Union[str, int]
    ) -> Document:
        """Load a specific document by its ID.

        Args:
            group_login: The group/user login name
            book_slug: The book identifier
            document_id: The specific document ID to load

        Returns:
            Iterator yielding a single Document object

        Yields:
            Document: The parsed document
        """
        try:
            # Get document detail directly using the API
            doc_detail = self.get_doc_detail(group_login, book_slug, document_id)
            return self.parse_document(doc_detail)

        except Exception as e:
            raise ValueError(f"Failed to load document {document_id}: {str(e)}")



from typing import List

from whiskerrag_types.interface.loader_interface import BaseLoader
from whiskerrag_types.model.knowledge import KnowledgeSourceEnum, YuqueSourceConfig
from whiskerrag_types.model.multi_modal import Text
from whiskerrag_utils.helper.yuque import ExtendedYuqueLoader
from whiskerrag_utils.registry import RegisterTypeEnum, register


@register(RegisterTypeEnum.KNOWLEDGE_LOADER, KnowledgeSourceEnum.YUQUE)
class WhiskerYuqueLoader(BaseLoader[Text]):

    async def load(self) -> List[Text]:
        if not isinstance(self.knowledge.source_config, YuqueSourceConfig):
            raise AttributeError("Invalid source config type for YuqueLoader")
        group_login = self.knowledge.source_config.group_login
        book_slug = self.knowledge.source_config.book_slug
        document_id = self.knowledge.source_config.document_id
        text_list: List[Text] = []
        try:
            loader = ExtendedYuqueLoader(
                access_token=self.knowledge.source_config.auth_info,
                api_url=self.knowledge.source_config.api_url,
            )
            # Extract book_id and document_id from source_config
            group_login = self.knowledge.source_config.group_login
            book_slug = self.knowledge.source_config.book_slug
            document_id = self.knowledge.source_config.document_id
            if not group_login:
                raise ValueError("group_login is needed for WhiskerYuqueLoader")
            if not book_slug:
                raise ValueError("book_slug is needed for WhiskerYuqueLoader")
            if not document_id:
                raise ValueError("document_id is needed for WhiskerYuqueLoader")

            try:
                parsed_document = loader.load_document_by_path(
                    group_login, book_slug, document_id
                )
                text_list.append(
                    Text(
                        content=parsed_document.page_content,
                        metadata=parsed_document.metadata,
                    )
                )

            except Exception as e:
                raise ValueError(f"Failed to get document: {e}")
            # Check if only book_id is provided
            return text_list

        except Exception as e:
            raise Exception(f"Failed to load content from Yuque: {e}")



import asyncio
from typing import List, Optional, Union

from whiskerrag_types.interface.embed_interface import BaseEmbedding
from whiskerrag_types.interface.parser_interface import ParseResult
from whiskerrag_types.model.chunk import Chunk
from whiskerrag_types.model.knowledge import Knowledge
from whiskerrag_types.model.multi_modal import Image, Text

from .registry import RegisterTypeEnum, get_register, init_register, register


async def process_parse_item(
    parse_item: Union[Text, Image],
    knowledge: Knowledge,
    EmbeddingCls: type[BaseEmbedding],
    semaphore: asyncio.Semaphore,
) -> Optional[Chunk]:
    async with semaphore:
        try:
            if isinstance(parse_item, Text):
                embedding = await EmbeddingCls().embed_text(
                    parse_item.content, timeout=30
                )
            elif isinstance(parse_item, Image):
                embedding = await EmbeddingCls().embed_image(parse_item, timeout=30)
            else:
                print(f"[warn]: illegal split item :{parse_item}")
                return None

            combined_metadata = {**knowledge.metadata}
            if isinstance(parse_item, Text) and parse_item.metadata:
                combined_metadata.update(parse_item.metadata)

            return Chunk(
                context=(
                    parse_item.content
                    if isinstance(parse_item, Text)
                    else parse_item.url
                ),
                enabled=knowledge.enabled,
                metadata=combined_metadata,
                embedding=embedding,
                knowledge_id=knowledge.knowledge_id,
                embedding_model_name=knowledge.embedding_model_name,
                space_id=knowledge.space_id,
                tenant_id=knowledge.tenant_id,
            )
        except Exception as e:
            print(f"Error processing parse item: {e}")
            return None


async def get_chunks_by_knowledge(knowledge: Knowledge) -> List[Chunk]:
    """
    Convert knowledge into vectorized chunks with controlled concurrency

    Args:
        knowledge (Knowledge): Knowledge object containing source type, split configuration,
                             embedding model and other information

    Returns:
        List[Chunk]: List of vectorized chunks

    Process flow:
    1. Get corresponding loader based on knowledge source type
    2. Get parser
    3. Get embedding model
    4. Load content as Text or Image
    5. Split content
    6. Vectorize each split content
    7. Generate final list of Chunk objects
    """
    parse_type = getattr(knowledge.split_config, "type", "base")
    ParserCls = get_register(RegisterTypeEnum.PARSER, parse_type)
    # dirty logic : thirdly platform
    if parse_type == "geagraph":
        await ParserCls().parse(knowledge, None)
        return []
    LoaderCls = get_register(RegisterTypeEnum.KNOWLEDGE_LOADER, knowledge.source_type)
    EmbeddingCls = get_register(
        RegisterTypeEnum.EMBEDDING, knowledge.embedding_model_name
    )
    contents = await LoaderCls(knowledge).load()
    parse_results: ParseResult = []
    for content in contents:
        split_result = await ParserCls().parse(knowledge, content)
        parse_results.extend(split_result)

    semaphore = asyncio.Semaphore(3)
    tasks = [
        process_parse_item(parse_item, knowledge, EmbeddingCls, semaphore)
        for parse_item in parse_results
    ]
    chunks = await asyncio.gather(*tasks)
    return [chunk for chunk in chunks if chunk is not None]


__all__ = [
    "get_register",
    "register",
    "RegisterTypeEnum",
    "init_register",
    "SplitterEnum",
    "get_chunks_by_knowledge",
]



import json
from typing import List

from langchain_text_splitters import RecursiveJsonSplitter

from whiskerrag_types.interface.parser_interface import BaseParser, ParseResult
from whiskerrag_types.model.knowledge import JSONSplitConfig, Knowledge
from whiskerrag_types.model.multi_modal import Text
from whiskerrag_utils.registry import RegisterTypeEnum, register


@register(RegisterTypeEnum.PARSER, "json")
class JSONSplitter(BaseParser[Text]):
    async def parse(
        self,
        knowledge: Knowledge,
        content: Text,
    ) -> ParseResult:
        """Splits JSON content into smaller chunks based on the provided configuration."""
        json_content = {}
        split_config = knowledge.split_config
        if not isinstance(split_config, JSONSplitConfig):
            raise TypeError("knowledge.split_config must be of type JSONSplitConfig")
        try:
            json_content = json.loads(content.content)
            if not isinstance(json_content, dict):
                raise ValueError("JSON content must be a dictionary.")
        except json.JSONDecodeError:
            raise ValueError("Invalid JSON content provided for splitting.")
        except ValueError as e:
            raise ValueError(f"Error processing JSON content: {str(e)}")
        splitter = RecursiveJsonSplitter(
            max_chunk_size=split_config.max_chunk_size,
            min_chunk_size=split_config.min_chunk_size,
        )
        split_texts = splitter.split_text(
            json_content, convert_lists=True, ensure_ascii=False
        )
        return [Text(content=text, metadata=content.metadata) for text in split_texts]

    async def batch_parse(
        self,
        knowledge: Knowledge,
        content_list: List[Text],
    ) -> List[ParseResult]:
        return [await self.parse(knowledge, content) for content in content_list]



import re
from typing import Dict, List

from langchain_text_splitters import RecursiveCharacterTextSplitter

from whiskerrag_types.interface.parser_interface import BaseParser, ParseResult
from whiskerrag_types.model.knowledge import Knowledge, MarkdownSplitConfig
from whiskerrag_types.model.multi_modal import Text
from whiskerrag_utils.registry import RegisterTypeEnum, register


@register(RegisterTypeEnum.PARSER, "markdown")
class MarkdownSplitter(BaseParser[Text]):
    async def parse(
        self,
        knowledge: Knowledge,
        content: Text,
    ) -> ParseResult:
        split_config = knowledge.split_config
        if not isinstance(split_config, MarkdownSplitConfig):
            raise TypeError(
                "knowledge.split_config must be of type MarkdownSplitConfig"
            )
        texts_to_process = (
            self._split_by_headers(content)
            if split_config.extract_header_first
            else [content]
        )
        final_chunks: ParseResult = []
        splitter = self._create_recursive_splitter(split_config)

        for text_item in texts_to_process:
            split_texts = splitter.split_text(text_item.content)
            chunks = [
                Text(content=text, metadata=text_item.metadata.copy())
                for text in split_texts
            ]
            final_chunks.extend(chunks)
        return final_chunks

    def _split_by_headers(self, content: Text) -> List[Text]:
        chunks = []
        current_chunk: List[str] = []
        current_headers: List[dict] = []

        lines = content.content.split("\n")

        for line in lines:
            header_match = re.match(r"^(#{1,6})\s+(.+)$", line)
            if header_match:
                if current_chunk:
                    chunks.append(
                        self._create_text_chunk(
                            content="\n".join(current_chunk),
                            headers=current_headers.copy(),
                            original_metadata=content.metadata,
                        )
                    )
                    current_chunk = []

                level = header_match.group(1)
                title = header_match.group(2)

                while current_headers and len(level) <= len(
                    current_headers[-1]["level"]
                ):
                    current_headers.pop()

                current_headers.append(
                    {"level": level, "title": title, "full_title": line.strip()}
                )

                current_chunk.append(line)
            else:
                current_chunk.append(line)

        if current_chunk:
            chunks.append(
                self._create_text_chunk(
                    content="\n".join(current_chunk),
                    headers=current_headers.copy(),
                    original_metadata=content.metadata,
                )
            )

        return chunks

    def _create_text_chunk(
        self, content: str, headers: List[Dict], original_metadata: Dict
    ) -> Text:
        metadata = original_metadata.copy()
        metadata.update(
            {
                "headers": headers,
                "level": len(headers[-1]["level"]) if headers else 0,
                "header_path": (
                    " > ".join(h["title"] for h in headers) if headers else ""
                ),
                "is_header_chunk": bool(headers),
            }
        )

        return Text(content=content, metadata=metadata)

    def _create_recursive_splitter(
        self, config: MarkdownSplitConfig
    ) -> RecursiveCharacterTextSplitter:
        if "" not in config.separators:
            config.separators.append("")
        return RecursiveCharacterTextSplitter(
            chunk_size=config.chunk_size,
            chunk_overlap=config.chunk_overlap,
            separators=config.separators,
            is_separator_regex=config.is_separator_regex,
            keep_separator=config.keep_separator or False,
        )

    async def batch_parse(
        self,
        knowledge: Knowledge,
        content_list: List[Text],
    ) -> List[ParseResult]:
        return [await self.parse(knowledge, content) for content in content_list]



from typing import List, Optional, Union

from deprecated import deprecated
from pydantic import BaseModel, ConfigDict, Field, field_serializer, field_validator

from whiskerrag_types.model.chunk import Chunk
from whiskerrag_types.model.knowledge import EmbeddingModelEnum


@deprecated("RetrievalConfig is deprecated, please use RetrievalRequest instead.")
class RetrievalBaseConfig(BaseModel):
    embedding_model_name: Union[EmbeddingModelEnum, str] = Field(
        ..., description="The name of the embedding model"
    )
    similarity_threshold: float = Field(
        0.5,
        ge=0.0,
        le=1.0,
        description="The similarity threshold, ranging from 0.0 to 1.0.",
    )
    top: int = Field(1024, ge=1, description="The maximum number of results to return.")
    metadata_filter: dict = Field({}, description="metadata filter")

    @field_serializer("embedding_model_name")
    def serialize_embedding_model_name(
        self, embedding_model_name: Optional[Union[EmbeddingModelEnum, str]]
    ) -> Optional[str]:
        if embedding_model_name:
            if isinstance(embedding_model_name, EmbeddingModelEnum):
                return embedding_model_name.value
            else:
                return embedding_model_name
        else:
            return None


@deprecated(
    "RetrievalBySpaceRequest is deprecated, please use RetrievalRequest instead."
)
class RetrievalBySpaceRequest(RetrievalBaseConfig):
    question: str = Field(..., description="The query question")
    space_id_list: List[str] = Field(..., description="space id list")


@deprecated(
    "RetrievalBySpaceRequest is deprecated, please use RetrievalRequest instead."
)
class RetrievalByKnowledgeRequest(RetrievalBaseConfig):
    question: str = Field(..., description="The query question")
    knowledge_id_list: List[str] = Field(..., description="knowledge id list")


class RetrievalConfig(BaseModel):
    type: str = Field(
        ...,
        description="The retrieval type. Each retrieval type corresponds to a specific retriever.",
    )
    model_config = ConfigDict(extra="allow")


class RetrievalRequest(BaseModel):
    content: str = Field(
        ...,
        description="The content to be searched, such as a question, text, or any other query string.",
    )
    image_url: Optional[str] = Field(default=None, description="image")
    config: RetrievalConfig = Field(
        ...,
        description="The configuration for the retrieval request. Must inherit from RetrievalConfig and have a 'type' attribute.",
    )

    @field_validator("content")
    def content_must_not_be_empty(cls, v: str) -> str:
        if not v.strip():
            raise ValueError("content must not be empty")
        return v


class RetrievalChunk(Chunk):
    similarity: float = Field(
        ..., description="The similarity of the chunk, ranging from 0.0 to 1.0."
    )
